#!/bin/bash -e

bindir=$(dirname $0)
source $bindir/functions

echo $0 $*
echo

function usage {
    echo "do_repo.sh REPO_DIR BRANCH TASK"
}

repo="$1"
branch="$2"
task="$3"

if [ -z "$repo" ]; then
    usage
    echo "Need to specify a repo!"
    exit 1
fi

if [ -z "$branch" ]; then
    usage
    echo "Need to specify a branch!"
    exit 1
fi

if [ -z "$task" ]; then
    usage
    echo "Need to specify a task!"
    exit 1
fi

commit_message="import zuul job settings from project-config

This is a mechanically generated patch to complete step 1 of moving
the zuul job settings out of project-config and into each project
repository.

Because there will be a separate patch on each branch, the branch
specifiers for branch-specific jobs have been removed.

Because this patch is generated by a script, there may be some
cosmetic changes to the layout of the YAML file(s) as the contents are
normalized.

See the python3-first goal document for details:
https://governance.openstack.org/tc/goals/stein/python3-first.html

Story: #2002586
Task: #$task

"

enable_tox

set -x

# NOTE(dhellmann): "git review -s" will just hang waiting for input if
# there is a problem setting up the gerrit remote. One cause of such
# trouble is a mis-match between the project setting within the
# .gitreview file and the actual repo name.
actual=$(cat $repo/.gitreview  | grep project | cut -f2 -d=)
expected="$(basename $(dirname $repo))/$(basename $repo)"
if [ "$actual" != "$expected" -a "$actual" != "${expected}.git" ]; then
    echo "WARNING: git review -s is likely to fail for $expected because .gitreview says $actual"
else
    (cd "$repo" && git review -s)
fi

new_branch=python3-first-$(basename $branch)

if (cd "$repo" && git branch | grep -q $new_branch); then
    echo "$new_branch already exists, reusing"
    (cd "$repo" && git checkout $new_branch)
else
    echo "creating $new_branch"
    (cd "$repo" &&
            git checkout -- . &&
            git clean -f -d)

    if ! (cd "$repo" && git checkout -q origin/$branch); then
        echo "Could not check out origin/$branch in $repo"
        exit 2
    fi

    (cd "$repo" && git checkout -b $new_branch)
fi


python3-first -v --debug jobs update --branch "$branch" "$repo"
RC=$?
if [ $RC -eq 2 ]; then
    echo "No changes"
    exit 1
fi
if [ $RC -ne 0 ]; then
    echo "FAIL"
    exit $RC
fi

# We have lots of git commands after this point so we can just change
# directories now.
pushd "$repo"

if ! git diff --ignore-all-space; then
    echo "No changes other than whitespace"
    git checkout -- .
    exit 1
fi

git add .
# NOTE(dhellmann): Some repositories have '.*' excluded by default so
# adding a new file requires a force flag.
if [ -f ".zuul.yaml" ]; then
    git add -f .zuul.yaml
fi
git commit -m "$commit_message" || exit 2

git show
